/*!
  @par Revision history:
  - 20.12.2005, c
*/
#ifndef _TERMS_H_
#define _TERMS_H_

#include "common.h"
BEGIN_C_DECLS

#include "fmfield.h"
#include "geommech.h"

#include "terms_adj_navier_stokes.h"
#include "terms_basic.h"
#include "terms_biot.h"
#include "terms_electric.h"
#include "terms_hyperelastic.h"
#include "terms_diffusion.h"
#include "terms_elastic.h"
#include "terms_navier_stokes.h"
#include "terms_piezo.h"
#include "terms_surface.h"
#include "terms_volume.h"
#include "terms_dot.h"
#include "terms_op.h"

END_C_DECLS

#endif /* Header */
