.. This (-*- rst -*-) format file contains commonly used link targets and name
   substitutions. It may be included in many files, therefore it should only
   contain link targets and name substitutions. Try grepping for "^\.\. _" to
   find plausible candidates for this list.

.. NOTE: reST targets are __not_case_sensitive__, so only one target definition
   is needed for sfepy, SfePy, Sfepy, etc...

.. sfepy links
.. _sfepy: https://sfepy.org
.. _sfepy web site: https://sfepy.org
.. _mailing list: https://mail.python.org/mm3/mailman3/lists/sfepy.python.org/
.. _sfepy mailing list: https://mail.python.org/mm3/mailman3/lists/sfepy.python.org/
.. _our mailing list: https://mail.python.org/mm3/mailman3/lists/sfepy.python.org/
.. _sfepy issues: https://github.com/sfepy/sfepy/issues
.. _issues: https://github.com/sfepy/sfepy/issues
.. _sfepy download: https://sfepy.org/doc-devel/downloads.html
.. _download: https://sfepy.org/doc-devel/downloads.html
.. _sfepy git repository: https://github.com/sfepy/sfepy
.. _sfepy/sfepy: https://github.com/sfepy/sfepy
.. _sfepy tarball: https://github.com/sfepy/sfepy/tarball/master
.. _our wiki: https://github.com/sfepy/sfepy/wiki
.. _gallery: https://sfepy.org/gallery

.. other links

.. _python: http://python.org
.. _pep-0008: http://www.python.org/dev/peps/pep-0008

.. _SciPy: http://www.scipy.org
.. _NumPy: http://www.numpy.org
.. _scikit-umfpack: https://github.com/scikit-umfpack/scikit-umfpack
.. _sparse matrix: http://docs.scipy.org/doc/scipy/reference/sparse.html
.. _numpy documentation guide: https://github.com/numpy/numpy/blob/master/doc/HOWTO_DOCUMENT.rst.txt
.. _docstring example: https://github.com/numpy/numpy/blob/master/doc/example.py
.. _numpydoc: http://pypi.python.org/pypi/numpydoc
.. _meshio: https://github.com/nschloe/meshio

.. _anaconda: https://store.continuum.io/cshop/anaconda/
.. _continuum analytics: http://www.continuum.io/
.. _conda-forge: https://conda-forge.github.io/
.. _Pre-built *SfePy* packages: https://anaconda.org/conda-forge/sfepy
.. _sfepy-feedstock: https://github.com/conda-forge/sfepy-feedstock
.. _enthought canopy: https://www.enthought.com/products/canopy/
.. _epd: http://www.enthought.com/products/epd.php
.. _WinPython: https://winpython.github.io/

.. _cython: http://cython.org
.. _pytables: http://www.pytables.org
.. _sympy: http://sympy.org
.. _ipython: http://ipython.org
.. _pyparsing: http://sourceforge.net/projects/pyparsing
.. _umfpack: http://faculty.cse.tamu.edu/davis/suitesparse.html
.. _umfpackpy: http://code.google.com/p/umfpackpy
.. _gmsh: http://gmsh.info
.. _gmsh reference manual: http://gmsh.info/doc/texinfo/gmsh.html
.. _medit: http://www.ann.jussieu.fr/~frey/software.html
.. _python mesh viewer: http://labs.freehackers.org/projects/pythonmeshviewer
.. _mayavi: http://code.enthought.com/projects/mayavi
.. _matplotlib: http://matplotlib.sourceforge.net
.. _igakit: https://bitbucket.org/dalcinl/igakit
.. _petsc4py: https://bitbucket.org/petsc/petsc4py
.. _PETSc: https://www.mcs.anl.gov/petsc
.. _slepc4py: https://bitbucket.org/slepc/slepc4py
.. _SLEPc: http://slepc.upv.es
.. _mpi4py: https://bitbucket.org/mpi4py/mpi4py
.. _Metis: http://glaros.dtc.umn.edu/gkhome/views/metis
.. _pymetis: http://mathema.tician.de/software/pymetis
.. _BDDCML: http://users.math.cas.cz/sistek/software/bddcml.html
.. _wxPython: https://wxpython.org/
.. _psutil: https://github.com/giampaolo/psutil
.. _PyVista: https://docs.pyvista.org

.. _git: http://git-scm.com
.. _github: http://github.com
.. _create a github account: https://github.com/signup/free
.. _gist: https://gist.github.com
.. _gitwash: http://github.com/matthew-brett/gitwash
.. _numpy gitwash: http://docs.scipy.org/doc/numpy/dev/gitwash
.. _tbaggery.com: http://tbaggery.com/2008/04/19/a-note-about-git-commit-messages.html
.. _msysgit: https://git-for-windows.github.io/

.. _sphinx: http://sphinx.pocoo.org
.. _restructuredtext primer: http://sphinx.pocoo.org/rest.html
.. _sphinx markup constructs: http://sphinx.pocoo.org/markup/index.html
.. _docutils restructuredtext: http://docutils.sourceforge.net/rst.html
.. _Read the Docs: https://readthedocs.org

.. _tex live: http://www.tug.org/texlive
.. _LaTeX: http://www.latex-project.org/

.. _Visual C++ Build Tools: http://landinghub.visualstudio.com/visual-cpp-build-tools
.. _Anaconda FAQ: https://docs.continuum.io/anaconda/faq#how-is-cpython-compiled
.. _mingw-w32/64: https://mingw-w64.org/doku.php

.. _new bsd license: http://www.opensource.org/licenses/bsd-license.php

.. _finite element method: http://en.wikipedia.org/wiki/Finite_element_method
.. _isogeometric analysis: http://en.wikipedia.org/wiki/Isogeometric_analysis
.. _salome download: http://www.salome-platform.org/downloads/current-version
.. _salome: http://www.salome-platform.org
.. _salome tutorial site: http://www.salome-platform.org/user-section/salome-tutorials
.. _freecad: https://www.freecadweb.org
.. _openscad: http://www.openscad.org
.. _python scripting in freecad: https://www.freecadweb.org/wiki/Python_scripting_tutorial
.. _netgen: https://ngsolve.org

.. _PyAMG Wiki: https://github.com/pyamg/pyamg/wiki
.. _PETSc Users Manual: http://www.mcs.anl.gov/petsc/petsc-current/docs/manual.pdf
.. _PETSc Documentation: http://www.mcs.anl.gov/petsc/petsc-current/docs
.. _MUMPS: http://mumps-solver.org
